# Copyright 2012 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import contextlib
import fixtures
import io
import sys
import testtools
from unittest import mock
import urllib
from urllib import parse as urlparse

from tackerclient.common import exceptions
from tackerclient import shell
from tackerclient.tacker import v1_0 as tackerV1_0
from tackerclient.tacker.v1_0 import TackerCommand
from tackerclient.tests.unit import test_utils
from tackerclient.v1_0 import client

API_VERSION = "1.0"
FORMAT = 'json'
TOKEN = 'testtoken'
ENDURL = 'localurl'


@contextlib.contextmanager
def capture_std_streams():
    fake_stdout, fake_stderr = io.StringIO(), io.StringIO()
    stdout, stderr = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = fake_stdout, fake_stderr
        yield fake_stdout, fake_stderr
    finally:
        sys.stdout, sys.stderr = stdout, stderr


class FakeStdout(io.IOBase):

    def __init__(self):
        self.content = []

    def write(self, text):
        self.content.append(text)

    def make_string(self):
        result = ''
        for line in self.content:
            result = result + line
        return result


class MyResp(object):
    def __init__(self, status_code, headers=None, reason=None):
        self.status_code = status_code
        self.headers = headers or {}
        self.reason = reason


class MyApp(object):
    def __init__(self, _stdout):
        self.stdout = _stdout


def end_url(path, query=None, format=FORMAT):
    _url_str = ENDURL + "/v" + API_VERSION + path + "." + format
    return query and _url_str + "?" + query or _url_str


class MyUrlComparator(object):
    def __init__(self, lhs, client):
        self.lhs = lhs
        self.client = client

    def equals(self, rhs):
        lhsp = urlparse.urlparse(self.lhs)
        rhsp = urlparse.urlparse(rhs)

        lhs_qs = urlparse.parse_qsl(lhsp.query)
        rhs_qs = urlparse.parse_qsl(rhsp.query)

        return (lhsp.scheme == rhsp.scheme and
                lhsp.netloc == rhsp.netloc and
                lhsp.path == rhsp.path and
                len(lhs_qs) == len(rhs_qs) and
                set(lhs_qs) == set(rhs_qs))

    def __str__(self):
        if self.client and self.client.format != FORMAT:
            lhs_parts = self.lhs.split("?", 1)
            if len(lhs_parts) == 2:
                lhs = ("%s.%s?%s" % (lhs_parts[0][:-4],
                                     self.client.format,
                                     lhs_parts[1]))
            else:
                lhs = ("%s.%s" % (lhs_parts[0][:-4],
                                  self.client.format))
            return lhs
        return self.lhs

    def __repr__(self):
        return str(self)

    def __eq__(self, rhs):
        return self.equals(rhs)

    def __ne__(self, rhs):
        return not self.__eq__(rhs)


class MyComparator(object):
    def __init__(self, lhs, client):
        self.lhs = lhs
        self.client = client

    def _com_dict(self, lhs, rhs):
        if len(lhs) != len(rhs):
            return False
        for key, value in lhs.items():
            if key not in rhs:
                return False
            rhs_value = rhs[key]
            if not self._com(value, rhs_value):
                return False
        return True

    def _com_list(self, lhs, rhs):
        if len(lhs) != len(rhs):
            return False
        for lhs_value in lhs:
            if lhs_value not in rhs:
                return False
        return True

    def _com(self, lhs, rhs):
        if lhs is None:
            return rhs is None
        if isinstance(lhs, dict):
            if not isinstance(rhs, dict):
                return False
            return self._com_dict(lhs, rhs)
        if isinstance(lhs, list):
            if not isinstance(rhs, list):
                return False
            return self._com_list(lhs, rhs)
        if isinstance(lhs, tuple):
            if not isinstance(rhs, tuple):
                return False
            return self._com_list(lhs, rhs)
        return lhs == rhs

    def equals(self, rhs):
        if self.client:
            rhs = self.client.deserialize(rhs, 200)
        return self._com(self.lhs, rhs)

    def __repr__(self):
        if self.client:
            return self.client.serialize(self.lhs)
        return str(self.lhs)

    def __eq__(self, rhs):
        return self.equals(rhs)

    def __ne__(self, rhs):
        return not self.__eq__(rhs)


class CLITestV10Base(testtools.TestCase):

    format = 'json'
    test_id = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
    id_field = 'id'

    def _find_resourceid(self, client, resource, name_or_id):
        return name_or_id

    def setUp(self, plurals={}):
        """Prepare the test environment."""
        super(CLITestV10Base, self).setUp()
        self.endurl = ENDURL
        self.fake_stdout = FakeStdout()
        self.useFixture(fixtures.MonkeyPatch('sys.stdout', self.fake_stdout))
        self.useFixture(fixtures.MonkeyPatch(
            'tackerclient.tacker.v1_0.find_resourceid_by_name_or_id',
            self._find_resourceid))
        self.useFixture(fixtures.MonkeyPatch(
            'tackerclient.tacker.v1_0.find_resourceid_by_id',
            self._find_resourceid))
        self.client = client.LegacyClient(token=TOKEN,
                                          endpoint_url=self.endurl)

    @mock.patch.object(TackerCommand, 'get_client')
    def _test_create_resource(self, resource, cmd,
                              name, myid, args,
                              position_names, position_values, mock_get,
                              tenant_id=None, get_client_called_count=1,
                              tags=None, admin_state_up=True, extra_body=None,
                              **kwargs):
        mock_get.return_value = self.client
        non_admin_status_resources = ['vim']
        if (resource in non_admin_status_resources):
            body = {resource: {}, }
        else:
            body = {resource: {'admin_state_up': admin_state_up, }, }
        if tenant_id:
            body[resource].update({'tenant_id': tenant_id})
        if tags:
            body[resource].update({'tags': tags})
        if extra_body:
            body[resource].update(extra_body)
        body[resource].update(kwargs)

        for i in range(len(position_names)):
            body[resource].update({position_names[i]: position_values[i]})
        ress = {resource:
                {self.id_field: myid}, }
        if name:
            ress[resource].update({'name': name})
        self.client.format = self.format
        resstr = self.client.serialize(ress)
        # url method body
        resource_plural = tackerV1_0._get_resource_plural(resource,
                                                          self.client)
        path = getattr(self.client, resource_plural + "_path")
        # Work around for LP #1217791. XML deserializer called from
        # MyComparator does not decodes XML string correctly.
        if self.format == 'json':
            _body = MyComparator(body, self.client)
            _content_type = 'application/json'
        else:
            _body = self.client.serialize(body)
            _content_type = 'application/zip'
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            mock_req.return_value = (MyResp(200), resstr)
            args.extend(['--request-format', self.format])
            cmd_parser = cmd.get_parser('create_' + resource)
            shell.run_command(cmd, cmd_parser, args)
            mock_req.assert_called_once_with(
                end_url(path, format=self.format), 'POST',
                body=_body,
                headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                content_type=_content_type)
        self.assertEqual(get_client_called_count, mock_get.call_count)
        _str = self.fake_stdout.make_string()
        self.assertIn(myid, _str)
        if name:
            self.assertIn(name, _str)

    @mock.patch.object(TackerCommand, 'get_client')
    def _test_list_columns(self, cmd, resources_collection,
                           resources_out, mock_get,
                           args=['-f', 'json']):
        mock_get.return_value = self.client
        self.client.format = self.format
        resstr = self.client.serialize(resources_out)

        path = getattr(self.client, resources_collection + "_path")
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            mock_req.return_value = (MyResp(200), resstr)
            args.extend(['--request-format', self.format])
            cmd_parser = cmd.get_parser("list_" + resources_collection)
            shell.run_command(cmd, cmd_parser, args)
            mock_req.assert_called_once_with(
                end_url(path, format=self.format), 'GET',
                body=None,
                headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                content_type='application/json')
        mock_get.assert_called_once_with()

    def _test_list_resources(self, resources, cmd, detail=False, tags=[],
                             fields_1=[], fields_2=[], page_size=None,
                             sort_key=[], sort_dir=[], response_contents=None,
                             base_args=None, path=None,
                             template_source=None):
        if response_contents is None:
            contents = [{self.id_field: 'myid1', },
                        {self.id_field: 'myid2', }, ]
        else:
            contents = response_contents
        reses = {resources: contents}
        self.client.format = self.format
        resstr = self.client.serialize(reses)
        # url method body
        query = ""
        args = base_args if base_args is not None else []
        if detail:
            args.append('-D')
        args.extend(['--request-format', self.format])
        if fields_1:
            for field in fields_1:
                args.append('--fields')
                args.append(field)
        if template_source is not None:
            args.append("--template-source")
            args.append(template_source)
            query += 'template_source=' + template_source

        if tags:
            args.append('--')
            args.append("--tag")
        for tag in tags:
            args.append(tag)
            if isinstance(tag, str):
                tag = urllib.quote(tag.encode('utf-8'))
            if query:
                query += "&tag=" + tag
            else:
                query = "tag=" + tag
        if (not tags) and fields_2:
            args.append('--')
        if fields_2:
            args.append("--fields")
            for field in fields_2:
                args.append(field)
        if detail:
            query = query and query + '&verbose=True' or 'verbose=True'
        fields_1.extend(fields_2)
        for field in fields_1:
            if query:
                query += "&fields=" + field
            else:
                query = "fields=" + field
        if page_size:
            args.append("--page-size")
            args.append(str(page_size))
            if query:
                query += "&limit=%s" % page_size
            else:
                query = "limit=%s" % page_size
        if sort_key:
            for key in sort_key:
                args.append('--sort-key')
                args.append(key)
                if query:
                    query += '&'
                query += 'sort_key=%s' % key
        if sort_dir:
            len_diff = len(sort_key) - len(sort_dir)
            if len_diff > 0:
                sort_dir += ['asc'] * len_diff
            elif len_diff < 0:
                sort_dir = sort_dir[:len(sort_key)]
            for dir in sort_dir:
                args.append('--sort-dir')
                args.append(dir)
                if query:
                    query += '&'
                query += 'sort_dir=%s' % dir
        if path is None:
            path = getattr(self.client, resources + "_path")
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            mock_req.return_value = (MyResp(200), resstr)
            with mock.patch.object(TackerCommand, 'get_client') as mock_get:
                mock_get.return_value = self.client
                cmd_parser = cmd.get_parser("list_" + resources)
                shell.run_command(cmd, cmd_parser, args)
                mock_req.assert_called_once_with(
                    MyUrlComparator(end_url(path, query, format=self.format),
                                    self.client),
                    'GET',
                    body=None,
                    headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                    content_type='application/json')
        _str = self.fake_stdout.make_string()
        if response_contents is None:
            self.assertIn('myid1', _str)
        return _str

    @mock.patch.object(TackerCommand, 'get_client')
    def _test_list_sub_resources(self, resources, api_resource, cmd, myid,
                                 mock_get, detail=False,
                                 tags=[], fields_1=[], fields_2=[],
                                 page_size=None, sort_key=[], sort_dir=[],
                                 response_contents=None, base_args=None,
                                 path=None):
        mock_get.return_value = self.client
        if response_contents is None:
            contents = [{self.id_field: 'myid1', },
                        {self.id_field: 'myid2', }, ]
        else:
            contents = response_contents
        reses = {api_resource: contents}
        self.client.format = self.format
        resstr = self.client.serialize(reses)
        # url method body
        query = ""
        args = base_args if base_args is not None else []
        if detail:
            args.append('-D')
        args.extend(['--request-format', self.format])
        if fields_1:
            for field in fields_1:
                args.append('--fields')
                args.append(field)

        if tags:
            args.append('--')
            args.append("--tag")
        for tag in tags:
            args.append(tag)
            if isinstance(tag, str):
                tag = urllib.quote(tag.encode('utf-8'))
            if query:
                query += "&tag=" + tag
            else:
                query = "tag=" + tag
        if (not tags) and fields_2:
            args.append('--')
        if fields_2:
            args.append("--fields")
            for field in fields_2:
                args.append(field)
        if detail:
            query = query and query + '&verbose=True' or 'verbose=True'
        fields_1.extend(fields_2)
        for field in fields_1:
            if query:
                query += "&fields=" + field
            else:
                query = "fields=" + field
        if page_size:
            args.append("--page-size")
            args.append(str(page_size))
            if query:
                query += "&limit=%s" % page_size
            else:
                query = "limit=%s" % page_size
        if sort_key:
            for key in sort_key:
                args.append('--sort-key')
                args.append(key)
                if query:
                    query += '&'
                query += 'sort_key=%s' % key
        if sort_dir:
            len_diff = len(sort_key) - len(sort_dir)
            if len_diff > 0:
                sort_dir += ['asc'] * len_diff
            elif len_diff < 0:
                sort_dir = sort_dir[:len(sort_key)]
            for dir in sort_dir:
                args.append('--sort-dir')
                args.append(dir)
                if query:
                    query += '&'
                query += 'sort_dir=%s' % dir
        if path is None:
            path = getattr(self.client, resources + "_path")
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            mock_req.return_value = (MyResp(200), resstr)
            comparator = MyUrlComparator(
                end_url(path % myid, query=query, format=self.format),
                self.client)
            args.extend(['--request-format', self.format])
            cmd_parser = cmd.get_parser("list_" + resources)
            shell.run_command(cmd, cmd_parser, args)
            mock_req.assert_called_once_with(
                comparator, 'GET',
                body=None,
                headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                content_type='application/json')
        _str = self.fake_stdout.make_string()
        if response_contents is None:
            self.assertIn('myid1', _str)
        return _str

    # TODO(gongysh) add pagination unit test BUG 1633255
    # def _test_list_sub_resources_with_pagination(
    #      self, resources, api_resource, cmd, myid):
    #     self.mox.StubOutWithMock(cmd, "get_client")
    #     self.mox.StubOutWithMock(self.client.httpclient, "request")
    #     cmd.get_client().MultipleTimes().AndReturn(self.client)
    #     path = getattr(self.client, resources + "_path")
    #     fake_query = "marker=myid2&limit=2"
    #     reses1 = {api_resource: [{'id': 'myid1', },
    #                              {'id': 'myid2', }],
    #               '%s_links' % api_resource: [
    #                   {'href': end_url(path % myid, fake_query),
    #                    'rel': 'next'}]
    #               }
    #     reses2 = {api_resource: [{'id': 'myid3', },
    #                              {'id': 'myid4', }]}
    #     self.client.format = self.format
    #     resstr1 = self.client.serialize(reses1)
    #     resstr2 = self.client.serialize(reses2)
    #     self.client.httpclient.request(
    #         end_url(path % myid, "", format=self.format), 'GET',
    #         body=None,
    #         headers=mox.ContainsKeyValue(
    #             'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr1))
    #     self.client.httpclient.request(
    #         MyUrlComparator(end_url(path % myid, fake_query,
    #                                 format=self.format), self.client), 'GET',
    #         body=None, headers=mox.ContainsKeyValue(
    #             'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr2))
    #     self.mox.ReplayAll()
    #     cmd_parser = cmd.get_parser("list_" + resources)
    #     args = [myid, '--request-format', self.format]
    #     shell.run_command(cmd, cmd_parser, args)
    #     self.mox.VerifyAll()
    #     self.mox.UnsetStubs()

    # def _test_list_resources_with_pagination(self, resources, cmd):
    #     self.mox.StubOutWithMock(cmd, "get_client")
    #     self.mox.StubOutWithMock(self.client.httpclient, "request")
    #     cmd.get_client().MultipleTimes().AndReturn(self.client)
    #     path = getattr(self.client, resources + "_path")
    #     fake_query = "marker=myid2&limit=2"
    #     reses1 = {resources: [{'id': 'myid1', },
    #                           {'id': 'myid2', }],
    #               '%s_links' % resources: [
    #                   {'href': end_url(path, fake_query),
    #                    'rel': 'next'}]}
    #     reses2 = {resources: [{'id': 'myid3', },
    #                           {'id': 'myid4', }]}
    #     self.client.format = self.format
    #     resstr1 = self.client.serialize(reses1)
    #     resstr2 = self.client.serialize(reses2)
    #     self.client.httpclient.request(
    #         end_url(path, "", format=self.format), 'GET',
    #         body=None,
    #         headers=mox.ContainsKeyValue(
    #             'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr1))
    #     self.client.httpclient.request(
    #         MyUrlComparator(end_url(path, fake_query, format=self.format),
    #                         self.client), 'GET', body=None,
    #         headers=mox.ContainsKeyValue(
    #             'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr2))
    #     self.mox.ReplayAll()
    #     cmd_parser = cmd.get_parser("list_" + resources)
    #     args = ['--request-format', self.format]
    #     shell.run_command(cmd, cmd_parser, args)
    #     self.mox.VerifyAll()
    #     self.mox.UnsetStubs()

    @mock.patch.object(TackerCommand, 'get_client')
    def _test_update_resource(self, resource, cmd, myid, args, extrafields,
                              mock_get, get_client_called_count=1):
        mock_get.return_value = self.client
        body = {resource: extrafields}
        path = getattr(self.client, resource + "_path")
        self.client.format = self.format
        # Work around for LP #1217791. XML deserializer called from
        # MyComparator does not decodes XML string correctly.
        if self.format == 'json':
            _body = MyComparator(body, self.client)
            _content_type = 'application/json'
        else:
            _body = self.client.serialize(body)
            _content_type = 'application/zip'
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            comparator = MyUrlComparator(
                end_url(path % myid, format=self.format), self.client)
            mock_req.return_value = (MyResp(204), None)
            args.extend(['--request-format', self.format])
            cmd_parser = cmd.get_parser("update_" + resource)
            shell.run_command(cmd, cmd_parser, args)
            mock_req.assert_called_once_with(
                comparator,
                'PUT',
                body=_body,
                headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                content_type=_content_type)

        self.assertEqual(get_client_called_count, mock_get.call_count)
        _str = self.fake_stdout.make_string()
        self.assertIn(myid, _str)

    def _test_show_resource(self, resource, cmd, myid, args, fields=[]):
        with mock.patch.object(cmd, 'get_client') as mock_get:
            mock_get.return_value = self.client
            query = "&".join(["fields=%s" % field for field in fields])
            expected_res = {resource:
                            {self.id_field: myid,
                             'name': 'myname', }, }
            self.client.format = self.format
            resstr = self.client.serialize(expected_res)
            path = getattr(self.client, resource + "_path")
            with mock.patch.object(self.client.httpclient, 'request') as\
                    mock_req:
                mock_req.return_value = (MyResp(200), resstr)
                args.extend(['--request-format', self.format])
                cmd_parser = cmd.get_parser("show_" + resource)
                shell.run_command(cmd, cmd_parser, args)
                mock_req.assert_called_once_with(
                    end_url(path % myid, query, format=self.format), 'GET',
                    body=None,
                    headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                    content_type='application/json')
            _str = self.fake_stdout.make_string()
            mock_get.assert_called_once_with()
            self.assertIn(myid, _str)
            self.assertIn('myname', _str)

    @mock.patch.object(TackerCommand, 'get_client')
    def _test_delete_resource(self, resource, cmd, myid, args, mock_get):
        deleted_msg = {'vnf': 'delete initiated'}
        mock_get.return_value = self.client
        path = getattr(self.client, resource + "_path")
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            mock_req.return_value = (MyResp(204), None)
            args.extend(['--request-format', self.format])
            cmd_parser = cmd.get_parser("delete_" + resource)
            shell.run_command(cmd, cmd_parser, args)
            if '--force' in args:
                body_str = '{"' + resource + \
                           '": {"attributes": {"force": true}}}'
                mock_req.assert_called_once_with(
                    end_url(path % myid, format=self.format), 'DELETE',
                    body=body_str,
                    headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                    content_type='application/json')
            else:
                mock_req.assert_called_once_with(
                    end_url(path % myid, format=self.format), 'DELETE',
                    body=None,
                    headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                    content_type='application/json')
        mock_get.assert_called_once_with()
        _str = self.fake_stdout.make_string()
        msg = 'All specified %(resource)s(s) %(msg)s successfully\n' % {
            'msg': deleted_msg.get(resource, 'deleted'),
            'resource': resource}
        self.assertEqual(msg, _str)

    @mock.patch.object(TackerCommand, 'get_client')
    def _test_update_resource_action(self, resource, cmd, myid, action, args,
                                     body, mock_get, retval=None):
        mock_get.return_value = self.client
        path = getattr(self.client, resource + "_path")
        path_action = '%s/%s' % (myid, action)
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            mock_req.return_value = (MyResp(204), retval)
            args.extend(['--request-format', self.format])
            cmd_parser = cmd.get_parser("delete_" + resource)
            shell.run_command(cmd, cmd_parser, args)
            mock_req.assert_called_once_with(
                end_url(path % path_action, format=self.format), 'PUT',
                body=MyComparator(body, self.client),
                headers=test_utils.ContainsKeyValue('X-Auth-Token', TOKEN),
                content_type='application/json')
        _str = self.fake_stdout.make_string()
        self.assertIn(myid, _str)


class ClientV1TestJson(CLITestV10Base):
    def test_do_request_unicode(self):
        self.client.format = self.format
        unicode_text = '\u7f51\u7edc'
        action = '/test'
        params = {'test': unicode_text}
        body = params
        expect_body = self.client.serialize(body)
        self.client.httpclient.auth_token = unicode_text
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            mock_req.return_value = (MyResp(200), expect_body)
            res_body = self.client.do_request('PUT', action, body=body,
                                              params=params)
            expected_uri = 'localurl/v1.0/test.json?test=%E7%BD%91%E7%BB%9C'
            mock_req.assert_called_with(
                expected_uri, 'PUT', body=expect_body,
                headers={'X-Auth-Token': unicode_text,
                         'User-Agent': 'python-tackerclient'},
                content_type='application/json')
        # test response with unicode
        self.assertEqual(res_body, body)

    def test_do_request_error_without_response_body(self):
        self.client.format = self.format
        params = {'test': 'value'}
        expect_query = urlparse.urlencode(params)
        self.client.httpclient.auth_token = 'token'
        with mock.patch.object(self.client.httpclient, 'request') as mock_req:
            mock_req.return_value = (MyResp(400, reason='An error'), '')
            self.client.httpclient.request(
                end_url('/test', query=expect_query, format=self.format),
                'PUT', body='',
                headers={'X-Auth-Token': 'token'}
            )
            error = self.assertRaises(exceptions.TackerClientException,
                                      self.client.do_request, 'PUT', '/test',
                                      body='', params=params)
            self.assertEqual("400-tackerFault", str(error))


class CLITestV10ExceptionHandler(CLITestV10Base):

    def _test_exception_handler_v10(
            self, expected_exception, status_code, expected_msg,
            error_type=None, error_msg=None, error_detail=None,
            error_content=None):
        if error_content is None:
            error_content = {'TackerError': {'type': error_type,
                                             'message': error_msg,
                                             'detail': error_detail}}

        e = self.assertRaises(expected_exception,
                              client.exception_handler_v10,
                              status_code, error_content)
        self.assertEqual(status_code, e.status_code)
        self.assertEqual(expected_exception.__name__,
                         e.__class__.__name__)

        if expected_msg is None:
            if error_detail:
                expected_msg = '\n'.join([error_msg, error_detail])
            else:
                expected_msg = error_msg
        self.assertEqual(expected_msg, e.message)

    def test_exception_handler_v10_unknown_error_to_per_code_exception(self):
        for status_code, client_exc in exceptions.HTTP_EXCEPTION_MAP.items():
            error_msg = 'Unknown error'
            error_detail = 'This is detail'
            self._test_exception_handler_v10(
                client_exc, status_code,
                error_msg + '\n' + error_detail,
                'UnknownError', error_msg, error_detail)

    def test_exception_handler_v10_tacker_unknown_status_code(self):
        error_msg = 'Unknown error'
        error_detail = 'This is detail'
        self._test_exception_handler_v10(
            exceptions.TackerClientException, 501,
            error_msg + '\n' + error_detail,
            'UnknownError', error_msg, error_detail)

    def test_exception_handler_v10_bad_tacker_error(self):
        error_content = {'TackerError': {'unknown_key': 'UNKNOWN'}}
        self._test_exception_handler_v10(
            exceptions.TackerClientException, 500,
            expected_msg={'unknown_key': 'UNKNOWN'},
            error_content=error_content)

    def test_exception_handler_v10_error_dict_contains_message(self):
        error_content = {'message': 'This is an error message'}
        self._test_exception_handler_v10(
            exceptions.TackerClientException, 500,
            expected_msg='500-tackerFault',
            error_content=error_content)

    def test_exception_handler_v10_error_dict_not_contain_message(self):
        error_content = 'tackerFault'
        expected_msg = '%s-%s' % (500, error_content)
        self._test_exception_handler_v10(
            exceptions.TackerClientException, 500,
            expected_msg=expected_msg,
            error_content=error_content)

    def test_exception_handler_v10_default_fallback(self):
        error_content = 'This is an error message'
        expected_msg = '%s-%s' % (500, error_content)
        self._test_exception_handler_v10(
            exceptions.TackerClientException, 500,
            expected_msg=expected_msg,
            error_content=error_content)

    def test_exception_handler_v10_tacker_etsi_error(self):
        """Test ETSI error response"""

        known_error_map = [
            ({
                "status": "status 1",
                "detail": "sample 1"
            }, 400),
            ({
                "status": "status 2",
                "detail": "sample 2"
            }, 404),
            ({
                "status": "status 3",
                "detail": "sample 3"
            }, 409)
        ]

        for error_content, status_code in known_error_map:
            self._test_exception_handler_v10(
                exceptions.TackerClientException, status_code,
                expected_msg=error_content['detail'],
                error_content=error_content)



# Copyright 2011, VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Borrowed from nova code base, more utilities will be added/borrowed as and
# when needed.

"""Utilities and helper functions."""

import argparse
import logging
import os

from oslo_utils import encodeutils
from oslo_utils import importutils

from tackerclient.common import exceptions
from tackerclient.i18n import _


def env(*vars, **kwargs):
    """Returns the first environment variable set.

    If none are non-empty, defaults to '' or keyword arg default.
    """
    for v in vars:
        value = os.environ.get(v)
        if value:
            return value
    return kwargs.get('default', '')


def get_client_class(api_name, version, version_map):
    """Returns the client class for the requested API version.

    :param api_name: the name of the API, e.g. 'compute', 'image', etc
    :param version: the requested API version
    :param version_map: a dict of client classes keyed by version
    :rtype: a client class for the requested API version
    """
    try:
        client_path = version_map[str(version)]
    except (KeyError, ValueError):
        msg = _("Invalid %(api_name)s client version '%(version)s'. must be "
                "one of: %(map_keys)s")
        msg = msg % {'api_name': api_name, 'version': version,
                     'map_keys': ', '.join(version_map.keys())}
        raise exceptions.UnsupportedVersion(message=msg)

    return importutils.import_class(client_path)


def get_item_properties(item, fields, mixed_case_fields=(), formatters=None):
    """Return a tuple containing the item properties.

    :param item: a single item resource (e.g. Server, Tenant, etc)
    :param fields: tuple of strings with the desired field names
    :param mixed_case_fields: tuple of field names to preserve case
    :param formatters: dictionary mapping field names to callables
       to format the values
    """
    if formatters is None:
        formatters = {}

    row = []

    for field in fields:
        if field in formatters:
            row.append(formatters[field](item))
        else:
            if field in mixed_case_fields:
                field_name = field.replace(' ', '_')
            else:
                field_name = field.lower().replace(' ', '_')
            if not hasattr(item, field_name) and isinstance(item, dict):
                data = item[field_name]
            else:
                data = getattr(item, field_name, '')
            if data is None:
                data = ''
            row.append(data)
    return tuple(row)


def str2bool(strbool):
    if strbool is None:
        return None
    return strbool.lower() == 'true'


def str2dict(strdict):
    """Convert key1=value1,key2=value2,... string into dictionary.

    :param strdict: key1=value1,key2=value2
    """
    if not strdict:
        return {}
    return dict([kv.split('=', 1) for kv in strdict.split(',')])


def http_log_req(_logger, args, kwargs):
    if not _logger.isEnabledFor(logging.DEBUG):
        return

    string_parts = ['curl -i']
    for element in args:
        if element in ('GET', 'POST', 'DELETE', 'PUT'):
            string_parts.append(' -X %s' % element)
        else:
            string_parts.append(' %s' % element)

    for element in kwargs['headers']:
        header = ' -H "%s: %s"' % (element, kwargs['headers'][element])
        string_parts.append(header)

    if 'body' in kwargs and kwargs['body']:
        string_parts.append(" -d '%s'" % (kwargs['body']))
    req = encodeutils.safe_encode("".join(string_parts))
    _logger.debug("\nREQ: %s\n", req)


def http_log_resp(_logger, resp, body):
    if not _logger.isEnabledFor(logging.DEBUG):
        return
    _logger.debug("RESP:%(code)s %(headers)s %(body)s\n",
                  {'code': resp.status_code,
                   'headers': resp.headers,
                   'body': body})


def _safe_encode_without_obj(data):
    if isinstance(data, str):
        return encodeutils.safe_encode(data)
    return data


def safe_encode_list(data):
    return list(map(_safe_encode_without_obj, data))


def safe_encode_dict(data):
    def _encode_item(item):
        k, v = item
        if isinstance(v, list):
            return (k, safe_encode_list(v))
        elif isinstance(v, dict):
            return (k, safe_encode_dict(v))
        return (k, _safe_encode_without_obj(v))

    return dict(list(map(_encode_item, data.items())))


def add_boolean_argument(parser, name, **kwargs):
    for keyword in ('metavar', 'choices'):
        kwargs.pop(keyword, None)
    default = kwargs.pop('default', argparse.SUPPRESS)
    parser.add_argument(
        name,
        metavar='{True,False}',
        choices=['True', 'true', 'False', 'false'],
        default=default,
        **kwargs)


def get_file_path(filename):
    file_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                             '../%s' % filename))
    return file_path



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from functools import reduce
from osc_lib.command import command
from osc_lib import utils

from tackerclient.common import exceptions
from tackerclient.i18n import _
from tackerclient.osc import sdk_utils
from tackerclient.osc import utils as tacker_osc_utils

LOG = logging.getLogger(__name__)

_FORMATTERS = {
    'objectInstanceIds': tacker_osc_utils.FormatComplexDataColumn,
    'subObjectInstanceIds': tacker_osc_utils.FormatComplexDataColumn,
    'criteria': tacker_osc_utils.FormatComplexDataColumn,
    'reports': tacker_osc_utils.FormatComplexDataColumn,
    '_links': tacker_osc_utils.FormatComplexDataColumn
}

_MIXED_CASE_FIELDS = (
    'objectType', 'objectInstanceIds', 'subObjectInstanceIds', 'callbackUri'
)

_MIXED_CASE_FIELDS_UPDATE = (
    'callbackUri'
)

_VNF_PM_JOB_ID = 'vnf_pm_job_id'


def _get_columns(vnfpm_job_obj, action=None):
    if action == 'update':
        column_map = {
            'callbackUri': 'Callback Uri'
        }
    else:
        column_map = {
            'id': 'ID',
            'objectType': 'Object Type',
            'objectInstanceIds': 'Object Instance Ids',
            'subObjectInstanceIds': 'Sub Object Instance Ids',
            'criteria': 'Criteria',
            'callbackUri': 'Callback Uri',
            'reports': 'Reports',
            '_links': 'Links'
        }

    if action == 'show':
        column_map.update(
            {'reports': 'Reports'}
        )

    return sdk_utils.get_osc_show_columns_for_sdk_resource(
        vnfpm_job_obj, column_map)


class CreateVnfPmJob(command.ShowOne):
    _description = _("Create a new VNF PM job")

    def get_parser(self, prog_name):
        parser = super(CreateVnfPmJob, self).get_parser(prog_name)
        parser.add_argument(
            'request_file',
            metavar="<param-file>",
            help=_('Specify create VNF PM job request '
                   'parameters in a json file.'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        vnf_pm_job = client.create_vnf_pm_job(
            tacker_osc_utils.jsonfile2body(parsed_args.request_file))
        display_columns, columns = _get_columns(vnf_pm_job)
        data = utils.get_item_properties(
            sdk_utils.DictModel(vnf_pm_job), columns,
            formatters=_FORMATTERS, mixed_case_fields=_MIXED_CASE_FIELDS)
        return (display_columns, data)


class ListVnfPmJob(command.Lister):
    _description = _("List VNF PM jobs")

    def get_parser(self, prog_name):
        LOG.debug('get_parser(%s)', prog_name)
        parser = super(ListVnfPmJob, self).get_parser(prog_name)
        parser.add_argument(
            "--filter",
            metavar="<filter>",
            help=_("Attribute-based-filtering parameters"),
        )
        fields_exclusive_group = parser.add_mutually_exclusive_group(
            required=False)
        fields_exclusive_group.add_argument(
            "--all_fields",
            action="store_true",
            default=False,
            help=_("Include all complex attributes in the response"),
        )
        fields_exclusive_group.add_argument(
            "--fields",
            metavar="fields",
            help=_("Complex attributes to be included into the response"),
        )
        fields_exclusive_group.add_argument(
            "--exclude_fields",
            metavar="exclude-fields",
            help=_("Complex attributes to be excluded from the response"),
        )
        parser.add_argument(
            "--exclude_default",
            action="store_true",
            default=False,
            help=_("Indicates to exclude all complex attributes"
                   " from the response. This argument can be used alone or"
                   " with --fields and --filter. For all other combinations"
                   " tacker server will throw bad request error"),
        )
        return parser

    def case_modify(self, field):
        return reduce(
            lambda x, y: x + (' ' if y.isupper() else '') + y, field).title()

    def get_attributes(self, extra_fields=None, all_fields=False,
                       exclude_fields=None, exclude_default=False):
        fields = ['id', 'objectType', '_links']
        complex_fields = [
            'objectInstanceIds',
            'subObjectInstanceIds',
            'criteria',
            'reports']
        simple_fields = ['callbackUri']

        if extra_fields:
            fields.extend(extra_fields)

        if exclude_fields:
            fields.extend([field for field in complex_fields
                           if field not in exclude_fields])
        if all_fields:
            fields.extend(complex_fields)
            fields.extend(simple_fields)

        if exclude_default:
            fields.extend(simple_fields)

        attrs = []
        for field in fields:
            if field == '_links':
                attrs.extend([(field, 'Links', tacker_osc_utils.LIST_BOTH)])
            else:
                attrs.extend([(field, self.case_modify(field),
                               tacker_osc_utils.LIST_BOTH)])

        return tuple(attrs)

    def take_action(self, parsed_args):
        _params = {}
        extra_fields = []
        exclude_fields = []
        all_fields = False
        exclude_default = False
        if parsed_args.filter:
            _params['filter'] = parsed_args.filter
        if parsed_args.fields:
            _params['fields'] = parsed_args.fields
            fields = parsed_args.fields.split(',')
            for field in fields:
                extra_fields.append(field.split('/')[0])
        if parsed_args.exclude_fields:
            _params['exclude_fields'] = parsed_args.exclude_fields
            fields = parsed_args.exclude_fields.split(',')
            exclude_fields.extend(fields)
        if parsed_args.exclude_default:
            _params['exclude_default'] = None
            exclude_default = True
        if parsed_args.all_fields:
            _params['all_fields'] = None
            all_fields = True

        client = self.app.client_manager.tackerclient
        data = client.list_vnf_pm_jobs(**_params)
        headers, columns = tacker_osc_utils.get_column_definitions(
            self.get_attributes(extra_fields, all_fields, exclude_fields,
                                exclude_default), long_listing=True)
        return (headers,
                (utils.get_dict_properties(
                    s, columns, formatters=_FORMATTERS,
                    mixed_case_fields=_MIXED_CASE_FIELDS,
                ) for s in data['vnf_pm_jobs']))


class ShowVnfPmJob(command.ShowOne):
    _description = _("Display VNF PM job details")

    def get_parser(self, prog_name):
        parser = super(ShowVnfPmJob, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_PM_JOB_ID,
            metavar="<vnf-pm-job-id>",
            help=_("VNF PM job ID to display"))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        obj = client.show_vnf_pm_job(parsed_args.vnf_pm_job_id)
        display_columns, columns = _get_columns(obj, action='show')
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj), columns,
            mixed_case_fields=_MIXED_CASE_FIELDS,
            formatters=_FORMATTERS)
        return (display_columns, data)


class UpdateVnfPmJob(command.ShowOne):
    _description = _("Update information about an individual VNF PM job")

    def get_parser(self, prog_name):
        LOG.debug('get_parser(%s)', prog_name)
        parser = super(UpdateVnfPmJob, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_PM_JOB_ID,
            metavar="<vnf-pm-job-id>",
            help=_("VNF PM job ID to update.")
        )
        parser.add_argument(
            'request_file',
            metavar="<param-file>",
            help=_('Specify update PM job request '
                   'parameters in a json file.'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        updated_values = client.update_vnf_pm_job(
            parsed_args.vnf_pm_job_id,
            tacker_osc_utils.jsonfile2body(parsed_args.request_file))
        display_columns, columns = _get_columns(updated_values, 'update')
        data = utils.get_item_properties(
            sdk_utils.DictModel(updated_values), columns,
            mixed_case_fields=_MIXED_CASE_FIELDS_UPDATE)
        return (display_columns, data)


class DeleteVnfPmJob(command.Command):
    _description = _("Delete VNF PM job")

    def get_parser(self, prog_name):
        parser = super(DeleteVnfPmJob, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_PM_JOB_ID,
            metavar="<vnf-pm-job-id>",
            nargs="+",
            help=_("VNF PM job ID(s) to delete"))
        return parser

    def take_action(self, parsed_args):
        error_count = 0
        client = self.app.client_manager.tackerclient
        vnf_pm_job_ids = parsed_args.vnf_pm_job_id

        for job_id in vnf_pm_job_ids:
            try:
                client.delete_vnf_pm_job(job_id)
            except Exception as e:
                error_count += 1
                LOG.error(_("Failed to delete VNF PM job with "
                            "ID '%(job_id)s': %(e)s"),
                          {'job_id': job_id, 'e': e})

        total = len(vnf_pm_job_ids)
        if error_count > 0:
            msg = (_("Failed to delete %(error_count)s of %(total)s "
                     "VNF PM jobs.") % {'error_count': error_count,
                                        'total': total})
            raise exceptions.CommandError(message=msg)

        if total > 1:
            print(_('All specified VNF PM jobs are deleted '
                    'successfully'))
        else:
            print(_("VNF PM job '%s' deleted "
                    "successfully") % vnf_pm_job_ids[0])



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_utils.fixture import uuidsentinel
from unittest import mock

from tackerclient.common import exceptions
from tackerclient.osc.v2.vnfpm import vnfpm_report
from tackerclient.tests.unit.osc import base
from tackerclient.tests.unit.osc.v1.fixture_data import client
from tackerclient.tests.unit.osc.v2 import vnfpm_report_fakes


class TestVnfPmReport(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture

    def setUp(self):
        super(TestVnfPmReport, self).setUp()
        self.url = client.TACKER_URL
        self.header = {'content-type': 'application/json'}
        self.app = mock.Mock()
        self.app_args = mock.Mock()
        self.client_manager = self.cs
        self.app.client_manager.tackerclient = self.client_manager


def _get_columns_vnfpm_report():
    columns = ['Entries']
    return columns


class TestShowVnfPmReport(TestVnfPmReport):

    def setUp(self):
        super(TestShowVnfPmReport, self).setUp()
        self.show_vnf_pm_reports = vnfpm_report.ShowVnfPmReport(
            self.app, self.app_args, cmd_name='vnfpm report show')

    def test_take_action(self):
        """Test of take_action()"""
        vnfpm_report_obj = vnfpm_report_fakes.vnf_pm_report_response()
        vnf_pm_job_id = uuidsentinel.vnf_pm_job_id
        vnf_pm_report_id = uuidsentinel.vnfpm_report_obj
        arg_list = [vnf_pm_job_id, vnf_pm_report_id]
        verify_list = [
            ('vnf_pm_job_id', vnf_pm_job_id),
            ('vnf_pm_report_id', vnf_pm_report_id)
        ]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_reports, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', vnf_pm_job_id,
            'reports', vnf_pm_report_id)

        self.requests_mock.register_uri(
            'GET', url, headers=self.header, json=vnfpm_report_obj)

        columns, data = (self.show_vnf_pm_reports.take_action(parsed_args))

        self.assertCountEqual(_get_columns_vnfpm_report(), columns)

        _, attributes = vnfpm_report._get_columns(vnfpm_report_obj)
        expected_data = vnfpm_report_fakes.get_vnfpm_report_data(
            vnfpm_report_obj, columns=attributes)

        print(f'123, {expected_data}')
        print(f'456, {data}')
        self.assertListItemsEqual(expected_data, data)

    def test_take_action_vnf_pm_report_id_not_found(self):
        """Test if vnf-pm-report-id does not find."""
        vnf_pm_job_id = uuidsentinel.vnf_pm_job_id
        vnf_pm_report_id = uuidsentinel.vnf_pm_report_id
        arg_list = [vnf_pm_job_id, vnf_pm_report_id]
        verify_list = [
            ('vnf_pm_job_id', vnf_pm_job_id),
            ('vnf_pm_report_id', vnf_pm_report_id)
        ]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_reports, arg_list, verify_list)

        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', vnf_pm_job_id,
            'reports', vnf_pm_report_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_pm_reports.take_action,
                          parsed_args)

    def test_take_action_internal_server_error(self):
        """Test for internal server error."""
        vnf_pm_job_id = uuidsentinel.vnf_pm_job_id
        vnf_pm_report_id = uuidsentinel.vnf_pm_report_id
        arg_list = [vnf_pm_job_id, vnf_pm_report_id]
        verify_list = [
            ('vnf_pm_job_id', vnf_pm_job_id),
            ('vnf_pm_report_id', vnf_pm_report_id)
        ]
        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_reports, arg_list, verify_list)

        url = os.path.join(
            self.url, 'vnfpm/v2/pm_jobs', vnf_pm_job_id,
            'reports', vnf_pm_report_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_pm_reports.take_action,
                          parsed_args)



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from osc_lib.command import command
from osc_lib import utils

from tackerclient.i18n import _
from tackerclient.osc import sdk_utils
from tackerclient.osc import utils as tacker_osc_utils

LOG = logging.getLogger(__name__)

_FORMATTERS = {
    'entries': tacker_osc_utils.FormatComplexDataColumn
}

_VNF_PM_JOB_ID = 'vnf_pm_job_id'
_VNF_PM_REPORT_ID = 'vnf_pm_report_id'


def _get_columns(vnfpm_report_obj):
    column_map = {
        'entries': 'Entries'
    }
    return sdk_utils.get_osc_show_columns_for_sdk_resource(
        vnfpm_report_obj, column_map)


class ShowVnfPmReport(command.ShowOne):
    _description = _("Display VNF PM report details")

    def get_parser(self, prog_name):
        parser = super(ShowVnfPmReport, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_PM_JOB_ID,
            metavar="<vnf-pm-job-id>",
            help=_("VNF PM job id where the VNF PM report is located"))
        parser.add_argument(
            _VNF_PM_REPORT_ID,
            metavar="<vnf-pm-report-id>",
            help=_("VNF PM report ID to display"))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        obj = client.show_vnf_pm_report(
            parsed_args.vnf_pm_job_id, parsed_args.vnf_pm_report_id)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj),
            columns, formatters=_FORMATTERS,
            mixed_case_fields=None)
        return (display_columns, data)



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from tackerclient.osc import utils as tacker_osc_utils


def create_vnf_fm_alarms(count=2):
    """Create multiple fake vnf packages.

    :param int count:
        The number of vnf_fm_alarms to fake
    :return:
        A list of fake vnf fm alarms dictionary
    """
    vnf_fm_alarms = []
    for i in range(0, count):
        unique_id = uuidutils.generate_uuid()
        vnf_fm_alarms.append(vnf_fm_alarm_response(attrs={'id': unique_id}))
    return vnf_fm_alarms


def vnf_fm_alarm_response(attrs=None, action=None):
    """Create a fake vnf fm alarm.

    :param Dictionary attrs:
        A dictionary with all attributes
    :return:
        A FakeVnfFmAlarm dict
    """

    if action == 'update':
        fake_vnf_fm_alarm = {
            "ackState": "UNACKNOWLEDGED"
        }
        return fake_vnf_fm_alarm

    attrs = attrs or {}
    # Set default attributes.
    fake_vnf_fm_alarm = {
        "id": "78a39661-60a8-4824-b989-88c1b0c3534a",
        "managedObjectId": "c61314d0-f583-4ab3-a457-46426bce02d3",
        "vnfcInstanceIds": "0e5f3086-4e79-47ed-a694-54c29155fa26",
        "rootCauseFaultyResource": {
            "faultyResource": {
                "vimConnectionId": "0d57e928-86a4-4445-a4bd-1634edae73f3",
                "resourceId": "4e6ccbe1-38ec-4b1b-a278-64de09ba01b3",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "faultyResourceType": "COMPUTE"
        },
        "alarmRaisedTime": "2021-09-03 10:21:03",
        "alarmChangedTime": "2021-09-04 10:21:03",
        "alarmClearedTime": "2021-09-05 10:21:03",
        "alarmAcknowledgedTime": "2021-09-06 10:21:03",
        "ackState": "UNACKNOWLEDGED",
        "perceivedSeverity": "WARNING",
        "eventTime": "2021-09-07 10:06:03",
        "eventType": "EQUIPMENT_ALARM",
        "faultType": "Fault Type",
        "probableCause": "The server cannot be connected.",
        "isRootCause": False,
        "correlatedAlarmIds": [
            "c88b624e-e997-4b17-b674-10ca2bab62e0",
            "c16d41fd-12e2-49a6-bb17-72faf702353f"
        ],
        "faultDetails": [
            "Fault",
            "Details"
        ],
        "_links": {
            "self": {
                "href": "/vnffm/v1/alarms/"
                        "78a39661-60a8-4824-b989-88c1b0c3534a"
            },
            "objectInstance": {
                "href": "/vnflcm/v1/vnf_instances/"
                        "0e5f3086-4e79-47ed-a694-54c29155fa26"
            }
        }
    }

    # Overwrite default attributes.
    fake_vnf_fm_alarm.update(attrs)

    return fake_vnf_fm_alarm


def get_vnffm_alarm_data(vnf_fm_alarm, columns=None):
    """Get the vnffm alarm.

    :return:
        A tuple object sorted based on the name of the columns.
    """
    complex_attributes = [
        'vnfcInstanceIds',
        'rootCauseFaultyResource',
        'correlatedAlarmIds',
        'faultDetails',
        '_links'
    ]

    for attribute in complex_attributes:
        if vnf_fm_alarm.get(attribute):
            vnf_fm_alarm.update(
                {attribute: tacker_osc_utils.FormatComplexDataColumn(
                    vnf_fm_alarm[attribute])})

    # return the list of data as per column order
    if columns:
        return tuple([vnf_fm_alarm[key] for key in columns])

    return tuple([vnf_fm_alarm[key] for key in sorted(
        vnf_fm_alarm.keys())])



# Copyright (C) 2023 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import os
import sys

from io import StringIO
from oslo_utils.fixture import uuidsentinel
from unittest import mock

from tackerclient import client as root_client
from tackerclient.common import exceptions
from tackerclient.osc import utils as tacker_osc_utils
from tackerclient.osc.v2.vnfpm import vnfpm_threshold
from tackerclient.tests.unit.osc import base
from tackerclient.tests.unit.osc.v1.fixture_data import client
from tackerclient.tests.unit.osc.v2 import vnfpm_threshold_fakes
from tackerclient.tests.unit.test_cli10 import MyResp
from tackerclient.v1_0 import client as proxy_client


class TestVnfPmThreshold(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture

    def setUp(self):
        super(TestVnfPmThreshold, self).setUp()
        self.url = client.TACKER_URL
        self.header = {'content-type': 'application/json'}
        self.app = mock.Mock()
        self.app_args = mock.Mock()
        self.client_manager = self.cs
        self.app.client_manager.tackerclient = self.client_manager


def _get_columns_vnfpm_threshold(action=None):
    if action == 'update':
        columns = ['Callback Uri']
    else:
        columns = ['ID', 'Object Type', 'Object Instance Id',
                   'Sub Object Instance Ids', 'Criteria', 'Callback Uri',
                   'Links']
    if action == 'list':
        columns = [
            'ID', 'Object Type', 'Links'
        ]
    return columns


@ddt.ddt
class TestCreateVnfPmThreshold(TestVnfPmThreshold):

    def setUp(self):
        super(TestCreateVnfPmThreshold, self).setUp()
        self.create_vnf_pm_threshold = vnfpm_threshold.CreateVnfPmThreshold(
            self.app, self.app_args, cmd_name='vnfpm threshold create')

    def test_create_no_args(self):
        self.assertRaises(base.ParserException, self.check_parser,
                          self.create_vnf_pm_threshold, [], [])

    @ddt.unpack
    def test_take_action(self):
        param_file = ("./tackerclient/osc/v2/vnfpm/samples/"
                      "create_vnf_pm_threshold_param_sample.json")

        arg_list = [param_file]
        verify_list = [('request_file', param_file)]

        parsed_args = self.check_parser(self.create_vnf_pm_threshold, arg_list,
                                        verify_list)

        response_json = vnfpm_threshold_fakes.vnf_pm_threshold_response()
        self.requests_mock.register_uri(
            'POST', os.path.join(self.url, 'vnfpm/v2/thresholds'),
            json=response_json, headers=self.header)

        actual_columns, data = (
            self.create_vnf_pm_threshold.take_action(parsed_args))
        self.assertCountEqual(_get_columns_vnfpm_threshold(),
                              actual_columns)

        _, attributes = vnfpm_threshold._get_columns(response_json)
        expected_data = vnfpm_threshold_fakes.get_vnfpm_threshold_data(
            response_json, columns=attributes)
        self.assertListItemsEqual(expected_data, data)


@ddt.ddt
class TestListVnfPmThreshold(TestVnfPmThreshold):

    def setUp(self):
        super(TestListVnfPmThreshold, self).setUp()
        self.list_vnf_pm_thresholds = vnfpm_threshold.ListVnfPmThreshold(
            self.app, self.app_args, cmd_name='vnfpm threshold list')

    def test_take_action(self):
        vnf_pm_threshold_objs = vnfpm_threshold_fakes.create_vnf_pm_thresholds(
            count=3)
        parsed_args = self.check_parser(self.list_vnf_pm_thresholds, [], [])
        self.requests_mock.register_uri(
            'GET',
            os.path.join(self.url, 'vnfpm/v2/thresholds'),
            json=vnf_pm_threshold_objs, headers=self.header)

        actual_columns, data = self.list_vnf_pm_thresholds.take_action(
            parsed_args)

        _, columns = tacker_osc_utils.get_column_definitions(
            vnfpm_threshold._ATTR_MAP, long_listing=True)

        expected_data = []
        for vnf_pm_threshold_obj_idx in vnf_pm_threshold_objs:
            expected_data.append(
                vnfpm_threshold_fakes.get_vnfpm_threshold_data(
                    vnf_pm_threshold_obj_idx, columns=columns))

        self.assertCountEqual(_get_columns_vnfpm_threshold(action='list'),
                              actual_columns)
        self.assertCountEqual(expected_data, list(data))

    def test_take_action_with_filter(self):
        vnf_pm_threshold_objs = vnfpm_threshold_fakes.create_vnf_pm_thresholds(
            count=3)
        parsed_args = self.check_parser(
            self.list_vnf_pm_thresholds,
            ["--filter", '(eq,perceivedSeverity,WARNING)'],
            [('filter', '(eq,perceivedSeverity,WARNING)')])
        self.requests_mock.register_uri(
            'GET', os.path.join(
                self.url,
                'vnfpm/v2/thresholds?filter=(eq,perceivedSeverity,WARNING)'),
            json=vnf_pm_threshold_objs, headers=self.header)

        actual_columns, data = self.list_vnf_pm_thresholds.take_action(
            parsed_args)

        _, columns = tacker_osc_utils.get_column_definitions(
            vnfpm_threshold._ATTR_MAP, long_listing=True)

        expected_data = []
        for vnf_pm_threshold_obj_idx in vnf_pm_threshold_objs:
            expected_data.append(
                vnfpm_threshold_fakes.get_vnfpm_threshold_data(
                    vnf_pm_threshold_obj_idx, columns=columns))

        self.assertCountEqual(_get_columns_vnfpm_threshold(action='list'),
                              actual_columns)
        self.assertListItemsEqual(expected_data, list(data))

    def test_take_action_with_incorrect_filter(self):
        parsed_args = self.check_parser(
            self.list_vnf_pm_thresholds,
            ["--filter", '(perceivedSeverity)'],
            [('filter', '(perceivedSeverity)')])

        url = os.path.join(
            self.url, 'vnfpm/v2/thresholds?filter=(perceivedSeverity)')

        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=400, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.list_vnf_pm_thresholds.take_action,
                          parsed_args)

    def test_take_action_internal_server_error(self):
        parsed_args = self.check_parser(
            self.list_vnf_pm_thresholds,
            ["--filter", '(eq,perceivedSeverity,WARNING)'],
            [('filter', '(eq,perceivedSeverity,WARNING)')])

        url = os.path.join(
            self.url,
            'vnfpm/v2/thresholds?filter=(eq,perceivedSeverity,WARNING)')

        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.list_vnf_pm_thresholds.take_action,
                          parsed_args)


class TestShowVnfPmThreshold(TestVnfPmThreshold):

    def setUp(self):
        super(TestShowVnfPmThreshold, self).setUp()
        self.show_vnf_pm_thresholds = vnfpm_threshold.ShowVnfPmThreshold(
            self.app, self.app_args, cmd_name='vnfpm threshold show')

    def test_take_action(self):
        vnfpm_threshold_obj = vnfpm_threshold_fakes.vnf_pm_threshold_response()

        arg_list = [vnfpm_threshold_obj['id']]
        verify_list = [('vnf_pm_threshold_id', vnfpm_threshold_obj['id'])]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_thresholds, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/thresholds', vnfpm_threshold_obj['id'])

        self.requests_mock.register_uri(
            'GET', url, headers=self.header, json=vnfpm_threshold_obj)

        columns, data = (self.show_vnf_pm_thresholds.take_action(parsed_args))

        self.assertCountEqual(_get_columns_vnfpm_threshold('show'), columns)

        _, attributes = vnfpm_threshold._get_columns(vnfpm_threshold_obj)
        self.assertListItemsEqual(
            vnfpm_threshold_fakes.get_vnfpm_threshold_data(
                vnfpm_threshold_obj, columns=attributes), data)

    def test_take_action_vnf_pm_threshold_id_not_found(self):
        arg_list = [uuidsentinel.vnf_pm_threshold_id]
        verify_list = [('vnf_pm_threshold_id',
                        uuidsentinel.vnf_pm_threshold_id)]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_thresholds, arg_list, verify_list)

        url = os.path.join(
            self.url, 'vnfpm/v2/thresholds',
            uuidsentinel.vnf_pm_threshold_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_pm_thresholds.take_action,
                          parsed_args)

    def test_take_action_internal_server_error(self):
        arg_list = [uuidsentinel.vnf_pm_threshold_id]
        verify_list = [('vnf_pm_threshold_id',
                        uuidsentinel.vnf_pm_threshold_id)]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_pm_thresholds, arg_list, verify_list)

        url = os.path.join(
            self.url, 'vnfpm/v2/thresholds',
            uuidsentinel.vnf_pm_threshold_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_pm_thresholds.take_action,
                          parsed_args)


@ddt.ddt
class TestUpdateVnfPmThreshold(TestVnfPmThreshold):

    def setUp(self):
        super(TestUpdateVnfPmThreshold, self).setUp()
        self.update_vnf_pm_threshold = vnfpm_threshold.UpdateVnfPmThreshold(
            self.app, self.app_args, cmd_name='vnfpm threshold update')

    def test_take_action(self):
        param_file = ("./tackerclient/osc/v2/vnfpm/samples/"
                      "update_vnf_pm_threshold_param_sample.json")
        arg_list = [uuidsentinel.vnf_pm_threshold_id, param_file]
        verify_list = [
            ('vnf_pm_threshold_id', uuidsentinel.vnf_pm_threshold_id),
            ('request_file', param_file)
        ]
        vnfpm_threshold_obj = vnfpm_threshold_fakes.vnf_pm_threshold_response(
            None, 'update')

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_pm_threshold, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/thresholds',
            uuidsentinel.vnf_pm_threshold_id)
        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, json=vnfpm_threshold_obj)

        actual_columns, data = (
            self.update_vnf_pm_threshold.take_action(parsed_args))
        expected_columns = _get_columns_vnfpm_threshold(action='update')
        self.assertCountEqual(expected_columns, actual_columns)

        _, columns = vnfpm_threshold._get_columns(
            vnfpm_threshold_obj, action='update')
        expected_data = vnfpm_threshold_fakes.get_vnfpm_threshold_data(
            vnfpm_threshold_obj, columns=columns)
        self.assertEqual(expected_data, data)

    @mock.patch.object(proxy_client.ClientBase, 'deserialize')
    def test_take_action_check_content_type(self, mock_des):
        param_file = ('./tackerclient/osc/v2/vnfpm/samples/'
                      'update_vnf_pm_threshold_param_sample.json')
        arg_list = [uuidsentinel.vnf_pm_threshold_id, param_file]
        verify_list = [
            ('vnf_pm_threshold_id', uuidsentinel.vnf_pm_threshold_id),
            ('request_file', param_file)
        ]
        vnfpm_threshold_obj = vnfpm_threshold_fakes.vnf_pm_threshold_response(
            None, 'update')
        mock_des.return_value = {}

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_pm_threshold, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/thresholds',
            uuidsentinel.vnf_pm_threshold_id)
        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, json=vnfpm_threshold_obj)

        with mock.patch.object(root_client.HTTPClient,
                               'do_request') as mock_req:
            headers = {'Content-Type': 'application/json'}
            mock_req.return_value = (MyResp(200, headers=headers), None)
            self.update_vnf_pm_threshold.take_action(parsed_args)
            mock_req.assert_called_once_with(
                f'/vnfpm/v2/thresholds/{uuidsentinel.vnf_pm_threshold_id}',
                'PATCH', body=mock.ANY, headers=mock.ANY,
                content_type='application/merge-patch+json', accept='json')

    def test_take_action_vnf_pm_threshold_id_not_found(self):
        param_file = ("./tackerclient/osc/v2/vnfpm/samples/"
                      "update_vnf_pm_threshold_param_sample.json")
        arg_list = [uuidsentinel.vnf_pm_threshold_id, param_file]
        verify_list = [
            ('vnf_pm_threshold_id', uuidsentinel.vnf_pm_threshold_id),
            ('request_file', param_file)
        ]

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_pm_threshold, arg_list, verify_list)
        url = os.path.join(
            self.url, 'vnfpm/v2/thresholds',
            uuidsentinel.vnf_pm_threshold_id)
        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, status_code=404, json={})
        self.assertRaises(exceptions.TackerClientException,
                          self.update_vnf_pm_threshold.take_action,
                          parsed_args)


class TestDeleteVnfPmThreshold(TestVnfPmThreshold):

    def setUp(self):
        super(TestDeleteVnfPmThreshold, self).setUp()
        self.delete_vnf_pm_threshold = vnfpm_threshold.DeleteVnfPmThreshold(
            self.app, self.app_args, cmd_name='vnfpm threshold delete')

        # Vnf Pm threshold to delete
        self.vnf_pm_thresholds = (
            vnfpm_threshold_fakes.create_vnf_pm_thresholds(count=3))

    def _mock_request_url_for_delete(self, index):
        url = os.path.join(self.url, 'vnfpm/v2/thresholds',
                           self.vnf_pm_thresholds[index]['id'])

        self.requests_mock.register_uri('DELETE', url,
                                        headers=self.header, json={})

    def test_delete_one_vnf_pm_threshold(self):
        arg_list = [self.vnf_pm_thresholds[0]['id']]
        verify_list = [('vnf_pm_threshold_id',
                       [self.vnf_pm_thresholds[0]['id']])]

        parsed_args = self.check_parser(self.delete_vnf_pm_threshold, arg_list,
                                        verify_list)

        self._mock_request_url_for_delete(0)
        sys.stdout = buffer = StringIO()
        result = self.delete_vnf_pm_threshold.take_action(parsed_args)
        self.assertIsNone(result)
        self.assertEqual(
            (f"VNF PM threshold '{self.vnf_pm_thresholds[0]['id']}' "
             f"deleted successfully"), buffer.getvalue().strip())

    def test_delete_multiple_vnf_pm_threshold(self):
        arg_list = []
        for obj in self.vnf_pm_thresholds:
            arg_list.append(obj['id'])
        verify_list = [('vnf_pm_threshold_id', arg_list)]
        parsed_args = self.check_parser(self.delete_vnf_pm_threshold, arg_list,
                                        verify_list)
        for i in range(0, len(self.vnf_pm_thresholds)):
            self._mock_request_url_for_delete(i)
        sys.stdout = buffer = StringIO()
        result = self.delete_vnf_pm_threshold.take_action(parsed_args)
        self.assertIsNone(result)
        self.assertEqual('All specified VNF PM thresholds are deleted '
                         'successfully', buffer.getvalue().strip())

    def test_delete_multiple_vnf_pm_threshold_exception(self):
        arg_list = [
            self.vnf_pm_thresholds[0]['id'],
            'xxxx-yyyy-zzzz',
            self.vnf_pm_thresholds[1]['id'],
        ]
        verify_list = [('vnf_pm_threshold_id', arg_list)]
        parsed_args = self.check_parser(self.delete_vnf_pm_threshold,
                                        arg_list, verify_list)

        self._mock_request_url_for_delete(0)

        url = os.path.join(self.url, 'vnfpm/v2/thresholds',
                           'xxxx-yyyy-zzzz')
        self.requests_mock.register_uri(
            'GET', url, exc=exceptions.ConnectionFailed)

        self._mock_request_url_for_delete(1)
        exception = self.assertRaises(exceptions.CommandError,
                                      self.delete_vnf_pm_threshold.take_action,
                                      parsed_args)

        self.assertEqual(
            f'Failed to delete 1 of {len(self.vnf_pm_thresholds)} '
            'VNF PM thresholds.', exception.message)



# Copyright 2011 VMware, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tackerclient.i18n import _

"""
Tacker base exception handling.

Exceptions are classified into three categories:
* Exceptions corresponding to exceptions from tacker server:
  This type of exceptions should inherit one of exceptions
  in HTTP_EXCEPTION_MAP.
* Exceptions from client library:
  This type of exceptions should inherit TackerClientException.
* Exceptions from CLI code:
  This type of exceptions should inherit TackerCLIError.
"""


class TackerException(Exception):
    """Base Tacker Exception.

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """
    message = _("An unknown exception occurred.")

    def __init__(self, message=None, **kwargs):
        if message:
            self.message = message
        try:
            self._error_string = self.message % kwargs
        except Exception:
            # at least get the core message out if something happened
            self._error_string = self.message

    def __str__(self):
        return self._error_string


class TackerClientException(TackerException):
    """Base exception which exceptions from Tacker are mapped into.

    NOTE: on the client side, we use different exception types in order
    to allow client library users to handle server exceptions in try...except
    blocks. The actual error message is the one generated on the server side.
    """

    status_code = 0

    def __init__(self, message=None, **kwargs):
        if 'status_code' in kwargs:
            self.status_code = kwargs['status_code']
        super(TackerClientException, self).__init__(message, **kwargs)


# Base exceptions from Tacker

class BadRequest(TackerClientException):
    status_code = 400


class Unauthorized(TackerClientException):
    status_code = 401
    message = _("Unauthorized: bad credentials.")


class Forbidden(TackerClientException):
    status_code = 403
    message = _("Forbidden: your credentials don't give you access to this "
                "resource.")


class NotFound(TackerClientException):
    status_code = 404


class Conflict(TackerClientException):
    status_code = 409


class InternalServerError(TackerClientException):
    status_code = 500


class ServiceUnavailable(TackerClientException):
    status_code = 503


HTTP_EXCEPTION_MAP = {
    400: BadRequest,
    401: Unauthorized,
    403: Forbidden,
    404: NotFound,
    409: Conflict,
    500: InternalServerError,
    503: ServiceUnavailable,
}


# Exceptions from client library

class NoAuthURLProvided(Unauthorized):
    message = _("auth_url was not provided to the Tacker client")


class EndpointNotFound(TackerClientException):
    message = _("Could not find Service or Region in Service Catalog.")


class EndpointTypeNotFound(TackerClientException):
    message = _("Could not find endpoint type %(type_)s in Service Catalog.")


class RequestURITooLong(TackerClientException):
    """Raised when a request fails with HTTP error 414."""

    def __init__(self, **kwargs):
        self.excess = kwargs.get('excess', 0)
        super(RequestURITooLong, self).__init__(**kwargs)


class ConnectionFailed(TackerClientException):
    message = _("Connection to tacker failed: %(reason)s")


class SslCertificateValidationError(TackerClientException):
    message = _("SSL certificate validation has failed: %(reason)s")


class MalformedResponseBody(TackerClientException):
    message = _("Malformed response body: %(reason)s")


class InvalidContentType(TackerClientException):
    message = _("Invalid content type %(content_type)s.")


class InvalidInput(TackerClientException):
    message = _("Invalid input: %(reason)s")


class EmptyInput(TackerClientException):
    message = _("Empty input: %(reason)s")


class UnsupportedCommandVersion(TackerClientException):
    message = _("This command is not supported in version %(version)s")


# Command line exceptions

class TackerCLIError(TackerException):
    """Exception raised when command line parsing fails."""
    pass


class CommandError(TackerCLIError):
    pass


class UnsupportedVersion(TackerCLIError):
    """Unsupported Version.

    Indicates that the user is trying to use an unsupported version of
    the API.
    """
    pass


class TackerClientNoUniqueMatch(TackerCLIError):
    message = _("Multiple %(resource)s matches found for name '%(name)s',"
                " use an ID to be more specific.")



# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# python-tackerclient documentation build configuration file

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#    sys.path.append(os.path.abspath('.'))

sys.path.insert(0, os.path.abspath('../..'))

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'reno.sphinxext',
    'openstackdocstheme',
    'cliff.sphinxext',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = 'OpenStack Contributors'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'openstackdocs'

# Output file base name for HTML help builder.
htmlhelp_basename = 'tackerclientdoc'

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.

# -- Options for manual page output -------------------------------------------

man_pages = [
    ('cli/index', 'tacker', 'Client for Tacker API',
     ['OpenStack Contributors'], 1),
]

# -- Options for openstackdocstheme -------------------------------------------

openstackdocs_repo_name = 'openstack/python-tackerclient'
openstackdocs_bug_project = 'python-tackerclient'
openstackdocs_bug_tag = 'doc'

# -- Options for cliff.sphinxext plugin ---------------------------------------

autoprogram_cliff_application = 'openstack'



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils.fixture import uuidsentinel
from oslo_utils import uuidutils

from tackerclient.osc import utils as tacker_osc_utils


def vnf_instance_response(attrs=None, instantiation_state='NOT_INSTANTIATED'):
    """Create a fake vnf instance.

    :param Dictionary attrs:
        A dictionary with all attributes
    :return:
        A vnf instance dict
    """
    attrs = attrs or {}

    # Set default attributes.
    dummy_vnf_instance = {
        "id": uuidsentinel.vnf_instance_id,
        "vnfInstanceName": "Fake-VNF-Instance",
        "vnfInstanceDescription": "Fake VNF",
        "vnfdId": uuidsentinel.vnf_package_vnfd_id,
        "vnfProvider": "NTT NS lab",
        "vnfProductName": "Sample VNF",
        "vnfSoftwareVersion": "1.0",
        "vnfdVersion": "1.0",
        "vnfPkgId": uuidsentinel.uuid,
        "_links": "vnflcm/v1/vnf_instances/" + uuidsentinel.vnf_instance_id +
                  "/instantiate",
        "instantiationState": instantiation_state,
        "vnfConfigurableProperties": {
            "test": "test_value"}}
    if instantiation_state == 'INSTANTIATED':
        dummy_vnf_instance.update({
            "vimConnectionInfo": [{
                'id': uuidsentinel.uuid,
                'vimId': uuidsentinel.vimId,
                'vimType': 'openstack',
                'interfaceInfo': {'k': 'v'},
                'accessInfo': {'k': 'v'},
                'extra': {'k': 'v'}
            }],
            "instantiatedVnfInfo": {
                "flavourId": uuidsentinel.flavourId,
                "vnfState": "STARTED",
                "extCpInfo": [{
                    'id': uuidsentinel.extCpInfo_uuid,
                    'cpdId': uuidsentinel.cpdId_uuid,
                    'cpProtocolInfo': [{
                        'layerProtocol': 'IP_OVER_ETHERNET',
                        'ipOverEthernet': '{}'
                    }],
                    'extLinkPortId': uuidsentinel.extLinkPortId_uuid,
                    'metadata': {'k': 'v'},
                    'associatedVnfcCpId': uuidsentinel.associatedVnfcCpId_uuid
                }],
                "extVirtualLinkInfo": [{
                    'id': uuidsentinel.extVirtualLinkInfo_uuid,
                    'resourceHandle': {},
                    'extLinkPorts': []
                }],
                "extManagedVirtualLinkInfo": [{
                    "id": uuidsentinel.extManagedVirtualLinkInfo_uuid,
                    'vnfVirtualLinkDescId': {},
                    'networkResource': {},
                    'vnfLinkPorts': []
                }],
                "vnfcResourceInfo": [{
                    'id': uuidsentinel.vnfcResourceInfo_uuid,
                    'vduId': uuidsentinel.vduId_uuid,
                    'computeResource': {},
                    'storageResourceIds': [],
                    'reservationId': uuidsentinel.reservationId,
                }],
                "vnfVirtualLinkResourceInfo": [{
                    'id': uuidsentinel.vnfVirtualLinkResourceInfo,
                    'vnfVirtualLinkDescId': 'VL4',
                    'networkResource': {},
                    'reservationId': uuidsentinel.reservationId,
                    'vnfLinkPorts': [],
                    'metadata': {'k': 'v'}
                }],
                "virtualStorageResourceInfo": [{
                    'id': uuidsentinel.virtualStorageResourceInfo,
                    'virtualStorageDescId': uuidsentinel.virtualStorageDescId,
                    'storageResource': {},
                    'reservationId': uuidsentinel.reservationId,
                    'metadata': {'k': 'v'}
                }]
            },
            "_links": {
                'self': 'self_link',
                'indicators': None,
                'instantiate': 'instantiate_link'
            }
        })

    # Overwrite default attributes.
    dummy_vnf_instance.update(attrs)

    return dummy_vnf_instance


def get_vnflcm_data(vnf_instance, list_action=False, columns=None):
    """Get the vnf instance data.

    :return:
        A tuple object sorted based on the name of the columns.
    """
    vnf = vnf_instance.copy()
    complex_attributes = ['vimConnectionInfo', 'instantiatedVnfInfo', '_links']
    for attribute in complex_attributes:
        if vnf.get(attribute):
            vnf.update(
                {attribute: tacker_osc_utils.FormatComplexDataColumn(
                    vnf[attribute])})

    if list_action:
        for item in ['vnfInstanceDescription', 'vnfdVersion']:
            vnf.pop(item)

    # return the list of data as per column order
    if columns:
        return tuple([vnf[key] for key in columns])

    return tuple([vnf[key] for key in sorted(
        vnf.keys())])


def create_vnf_instances(count=2):
    """Create multiple fake vnf instances.

    :param count: The number of vnf instances to fake
    :return:
        A list of fake vnf instances dictionary
    """
    vnf_instances = []
    for i in range(0, count):
        unique_id = uuidutils.generate_uuid()
        vnf_instances.append(vnf_instance_response(attrs={'id': unique_id}))
    return vnf_instances



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from osc_lib.command import command
from osc_lib import utils

from tackerclient.common import exceptions
from tackerclient.i18n import _
from tackerclient.osc import sdk_utils
from tackerclient.osc import utils as tacker_osc_utils

LOG = logging.getLogger(__name__)

_ATTR_MAP = (
    ('id', 'ID', tacker_osc_utils.LIST_BOTH),
    ('callbackUri', 'Callback Uri', tacker_osc_utils.LIST_BOTH)
)

_FORMATTERS = {
    'filter': tacker_osc_utils.FormatComplexDataColumn,
    '_links': tacker_osc_utils.FormatComplexDataColumn
}

_MIXED_CASE_FIELDS = (
    'callbackUri'
)

_VNF_FM_SUB_ID = 'vnf_fm_sub_id'


def _get_columns(vnffm_sub_obj):
    column_map = {
        'id': 'ID',
        'filter': 'Filter',
        'callbackUri': 'Callback Uri',
        '_links': 'Links'
    }

    return sdk_utils.get_osc_show_columns_for_sdk_resource(
        vnffm_sub_obj, column_map)


class CreateVnfFmSub(command.ShowOne):
    _description = _("Create a new VNF FM subscription")

    def get_parser(self, prog_name):
        parser = super(CreateVnfFmSub, self).get_parser(prog_name)
        parser.add_argument(
            'request_file',
            metavar="<param-file>",
            help=_('Specify create VNF FM subscription request '
                   'parameters in a json file.'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        vnf_fm_sub = client.create_vnf_fm_sub(
            tacker_osc_utils.jsonfile2body(parsed_args.request_file))
        display_columns, columns = _get_columns(vnf_fm_sub)
        data = utils.get_item_properties(
            sdk_utils.DictModel(vnf_fm_sub), columns,
            formatters=_FORMATTERS, mixed_case_fields=_MIXED_CASE_FIELDS)
        return (display_columns, data)


class ListVnfFmSub(command.Lister):
    _description = _("List VNF FM subs")

    def get_parser(self, prog_name):
        LOG.debug('get_parser(%s)', prog_name)
        parser = super(ListVnfFmSub, self).get_parser(prog_name)
        parser.add_argument(
            "--filter",
            metavar="<filter>",
            help=_("Attribute-based-filtering parameters"),
        )
        return parser

    def take_action(self, parsed_args):
        _params = {}
        if parsed_args.filter:
            _params['filter'] = parsed_args.filter

        client = self.app.client_manager.tackerclient
        data = client.list_vnf_fm_subs(**_params)
        headers, columns = tacker_osc_utils.get_column_definitions(
            _ATTR_MAP, long_listing=True)
        return (headers,
                (utils.get_dict_properties(
                    s, columns, formatters=_FORMATTERS,
                    mixed_case_fields=_MIXED_CASE_FIELDS,
                ) for s in data['vnf_fm_subs']))


class ShowVnfFmSub(command.ShowOne):
    _description = _("Display VNF FM subscription details")

    def get_parser(self, prog_name):
        parser = super(ShowVnfFmSub, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_FM_SUB_ID,
            metavar="<vnf-fm-sub-id>",
            help=_("VNF FM subscription ID to display"))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        obj = client.show_vnf_fm_sub(parsed_args.vnf_fm_sub_id)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj), columns,
            mixed_case_fields=_MIXED_CASE_FIELDS,
            formatters=_FORMATTERS)
        return (display_columns, data)


class DeleteVnfFmSub(command.Command):
    _description = _("Delete VNF FM subscription(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteVnfFmSub, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_FM_SUB_ID,
            metavar="<vnf-fm-sub-id>",
            nargs="+",
            help=_("VNF FM subscription ID(s) to delete"))
        return parser

    def take_action(self, parsed_args):
        error_count = 0
        client = self.app.client_manager.tackerclient
        vnf_fm_sub_ids = parsed_args.vnf_fm_sub_id

        for sub_id in vnf_fm_sub_ids:
            try:
                client.delete_vnf_fm_sub(sub_id)
            except Exception as e:
                error_count += 1
                LOG.error(_("Failed to delete VNF FM subscription with "
                            "ID '%(sub_id)s': %(e)s"),
                          {'sub_id': sub_id, 'e': e})

        total = len(vnf_fm_sub_ids)
        if error_count > 0:
            msg = (_("Failed to delete %(error_count)s of %(total)s "
                     "VNF FM subscriptions.") % {'error_count': error_count,
                                                 'total': total})
            raise exceptions.CommandError(message=msg)

        if total > 1:
            print(_('All specified VNF FM subscriptions are deleted '
                    'successfully'))
        else:
            print(_("VNF FM subscription '%s' deleted "
                    "successfully") % vnf_fm_sub_ids[0])



# Copyright (C) 2013 Yahoo! Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from tackerclient.common import exceptions
from tackerclient.common import utils


class TestUtils(testtools.TestCase):
    def test_string_to_bool_true(self):
        self.assertTrue(utils.str2bool('true'))

    def test_string_to_bool_false(self):
        self.assertFalse(utils.str2bool('false'))

    def test_string_to_bool_None(self):
        self.assertIsNone(utils.str2bool(None))

    def test_string_to_dictionary(self):
        input_str = 'key1=value1,key2=value2'
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(expected, utils.str2dict(input_str))

    def test_none_string_to_dictionary(self):
        input_str = ''
        expected = {}
        self.assertEqual(expected, utils.str2dict(input_str))
        input_str = None
        expected = {}
        self.assertEqual(expected, utils.str2dict(input_str))

    def test_get_dict_item_properties(self):
        item = {'name': 'test_name', 'id': 'test_id'}
        fields = ('name', 'id')
        actual = utils.get_item_properties(item=item, fields=fields)
        self.assertEqual(('test_name', 'test_id'), actual)

    def test_get_object_item_properties_mixed_case_fields(self):
        class Fake(object):
            def __init__(self):
                self.id = 'test_id'
                self.name = 'test_name'
                self.test_user = 'test'

        fields = ('name', 'id', 'test user')
        mixed_fields = ('test user', 'ID')
        item = Fake()
        actual = utils.get_item_properties(item, fields, mixed_fields)
        self.assertEqual(('test_name', 'test_id', 'test'), actual)

    def test_get_object_item_desired_fields_differ_from_item(self):
        class Fake(object):
            def __init__(self):
                self.id = 'test_id_1'
                self.name = 'test_name'
                self.test_user = 'test'

        fields = ('name', 'id', 'test user')
        item = Fake()
        actual = utils.get_item_properties(item, fields)
        self.assertNotEqual(('test_name', 'test_id', 'test'), actual)

    def test_get_object_item_desired_fields_is_empty(self):
        class Fake(object):
            def __init__(self):
                self.id = 'test_id_1'
                self.name = 'test_name'
                self.test_user = 'test'

        fields = []
        item = Fake()
        actual = utils.get_item_properties(item, fields)
        self.assertEqual((), actual)

    def test_get_object_item_with_formatters(self):
        class Fake(object):
            def __init__(self):
                self.id = 'test_id'
                self.name = 'test_name'
                self.test_user = 'test'

        class FakeCallable(object):
            def __call__(self, *args, **kwargs):
                return 'pass'

        fields = ('name', 'id', 'test user', 'is_public')
        formatters = {'is_public': FakeCallable()}
        item = Fake()
        act = utils.get_item_properties(item, fields, formatters=formatters)
        self.assertEqual(('test_name', 'test_id', 'test', 'pass'), act)


class ImportClassTestCase(testtools.TestCase):

    def test_get_client_class_invalid_version(self):
        self.assertRaises(
            exceptions.UnsupportedVersion,
            utils.get_client_class, 'image', '2', {'image': '2'})


class ContainsKeyValue(object):
    """Checks whether a key/value pair is in a dict parameter.

    The ContainsKeyValue class is a helper for mock.assert_*()
    method. It enables strict check than the built in mock.ANY
    helper, and is the equivalent of the mox.ContainsKeyValue()
    function from the legacy mox library

    Example usage could be:

      mock_some_method.assert_called_once_with(
            "hello",
            ContainsKeyValue('foo', bar),
            mock.ANY,
            "world",
            ContainsKeyValue('hello', world))
    """
    def __init__(self, wantkey, wantvalue):
        self.wantkey = wantkey
        self.wantvalue = wantvalue

    def __eq__(self, other):
        try:
            return other[self.wantkey] == self.wantvalue
        except (KeyError, TypeError):
            return False

    def __ne__(self, other):
        try:
            return other[self.wantkey] != self.wantvalue
        except (KeyError, TypeError):
            return True

    def __repr__(self):
        return "<ContainsKeyValue: key " + str(self.wantkey) + \
               " and value " + str(self.wantvalue) + ">"



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tackerclient.osc import utils as tacker_osc_utils


def vnf_pm_report_response(attrs=None):
    """Create a fake vnf pm report.

    :param Dictionary attrs:
        A dictionary with all attributes
    :return:
        A pm report dict
    """

    attrs = attrs or {}
    # Set default attributes.
    fake_vnf_pm_report = {
        "entries": [
            {
                "objectType": "VNFC",
                "objectInstanceId": "2bb72d78-b1d9-48fe-8c64-332654ffeb5d",
                "subObjectInstanceId": "09d46aed-3ec2-45d9-bfa2-add431e069b3",
                "performanceMetric":
                    "VCpuUsagePeakVnf.2bb72d78-b1d9-48fe-8c64-332654ffeb5d,",
                "performanceValues": [
                    {
                        "timeStamp": "2022/07/27 08:58:58",
                        "value": "1.88",
                        "context": {
                            "key": "value"
                        }
                    }
                ]
            }
        ]
    }

    # Overwrite default attributes.
    fake_vnf_pm_report.update(attrs)

    return fake_vnf_pm_report


def get_vnfpm_report_data(vnf_pm_report, columns=None):
    """Get the vnfpm report.

    :return:
        A tuple object sorted based on the name of the columns.
    """
    attribute = 'entries'

    if vnf_pm_report.get(attribute):
        vnf_pm_report.update(
            {attribute: tacker_osc_utils.FormatComplexDataColumn(
                vnf_pm_report[attribute])})

    # return the list of data as per column order
    if columns is None:
        columns = sorted(vnf_pm_report.keys())
    return tuple([vnf_pm_report[key] for key in columns])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import sys

from oslo_utils import encodeutils


def get_osc_show_columns_for_sdk_resource(
    sdk_resource,
    osc_column_map,
    invisible_columns=None
):
    """Get and filter the display and attribute columns for an SDK resource.

    Common utility function for preparing the output of an OSC show command.
    Some of the columns may need to get renamed, others made invisible.

    :param sdk_resource: An SDK resource
    :param osc_column_map: A hash of mappings for display column names
    :param invisible_columns: A list of invisible column names

    :returns: Two tuples containing the names of the display and attribute
              columns
    """

    if getattr(sdk_resource, 'allow_get', None) is not None:
        resource_dict = sdk_resource.to_dict(
            body=True, headers=False, ignore_none=False)
    else:
        resource_dict = sdk_resource

    # Build the OSC column names to display for the SDK resource.
    attr_map = {}
    display_columns = list(resource_dict.keys())
    invisible_columns = [] if invisible_columns is None else invisible_columns
    for col_name in invisible_columns:
        if col_name in display_columns:
            display_columns.remove(col_name)
    for sdk_attr, osc_attr in osc_column_map.items():
        if sdk_attr in display_columns:
            attr_map[osc_attr] = sdk_attr
            display_columns.remove(sdk_attr)
        if osc_attr not in display_columns:
            display_columns.append(osc_attr)
    sorted_display_columns = sorted(display_columns)

    # Build the SDK attribute names for the OSC column names.
    attr_columns = []
    for column in sorted_display_columns:
        new_column = attr_map[column] if column in attr_map else column
        attr_columns.append(new_column)
    return tuple(sorted_display_columns), tuple(attr_columns)


class DictModel(dict):
    """Convert dict into an object that provides attribute access to values."""

    def __init__(self, *args, **kwargs):
        """Convert dict values to DictModel values."""
        super(DictModel, self).__init__(*args, **kwargs)

        def needs_upgrade(item):
            return isinstance(item, dict) and not isinstance(item, DictModel)

        def upgrade(item):
            """Upgrade item if it needs to be upgraded."""
            if needs_upgrade(item):
                return DictModel(item)
            else:
                return item

        for key, value in self.items():
            if isinstance(value, (list, tuple)):
                # Keep the same type but convert dicts to DictModels
                self[key] = type(value)(
                    (upgrade(item) for item in value)
                )
            elif needs_upgrade(value):
                # Change dict instance values to DictModel instance values
                self[key] = DictModel(value)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    def __str__(self):
        pairs = ['%s=%s' % (k, v) for k, v in self.items()]
        return ', '.join(sorted(pairs))


def save_data(data, path):
    """Save data to the specified path.

    :param data: binary or string data
    :param path: file path to save data
    """
    if path is None:
        vnfpackage = getattr(sys.stdout, 'buffer', sys.stdout)
    else:
        mode = 'wb' if isinstance(data, bytes) else 'w'
        vnfpackage = open(path, mode)
    try:
        vnfpackage.write(data)
    finally:
        vnfpackage.close()


def exit(msg=None, exit_code=1):
    if msg:
        print(encodeutils.safe_decode(msg))
    sys.exit(exit_code)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import ddt
from unittest import mock

from tackerclient.common import exceptions
from tackerclient.osc.common.vnflcm import vnflcm_versions
from tackerclient.tests.unit.osc import base
from tackerclient.tests.unit.osc.v1.fixture_data import client


class TestVnfLcm(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture

    def setUp(self):
        super(TestVnfLcm, self).setUp()
        self.url = client.TACKER_URL
        self.header = {'content-type': 'application/json'}
        self.app = mock.Mock()
        self.app_args = mock.Mock()
        self.client_manager = self.cs
        self.app.client_manager.tackerclient = self.client_manager


@ddt.ddt
class TestVnfLcmVersions(TestVnfLcm):

    def setUp(self):
        super(TestVnfLcmVersions, self).setUp()
        self.vnflcm_versions = vnflcm_versions.VnfLcmVersions(
            self.app, self.app_args, cmd_name='vnflcm versions')

    def _versions_response(self, major_version=None):
        if major_version is None:
            return {"uriPrefix": "/vnflcm",
                    "apiVersions": [{"version": "1.3.0",
                                     "isDeprecated": False},
                                    {"version": "2.0.0",
                                     "isDeprecated": False}]}
        elif major_version == "1":
            return {"uriPrefix": "/vnflcm/v1",
                    "apiVersions": [{"version": "1.3.0",
                                     "isDeprecated": False}]}
        elif major_version == "2":
            return {"uriPrefix": "/vnflcm/v2",
                    "apiVersions": [{"version": "2.0.0",
                                     "isDeprecated": False}]}

    def test_invalid_major_version(self):
        parser = self.vnflcm_versions.get_parser('vnflcm versions')
        parsed_args = parser.parse_args(["--major-version", "3"])
        self.assertRaises(exceptions.InvalidInput,
                          self.vnflcm_versions.take_action,
                          parsed_args)

    def test_take_action_no_arg(self):
        parser = self.vnflcm_versions.get_parser('vnflcm versions')
        parsed_args = parser.parse_args([])

        response = self._versions_response()
        self.requests_mock.register_uri(
            'GET', os.path.join(self.url, 'vnflcm/api_versions'),
            json=response, headers=self.header)

        colmns, data = self.vnflcm_versions.take_action(parsed_args)

        self.assertEqual(colmns, tuple(response.keys()))
        self.assertEqual(data, tuple(response.values()))

    @ddt.data('1', '2')
    def test_take_action_with_major_version(self, major_version):
        parser = self.vnflcm_versions.get_parser('vnflcm versions')
        parsed_args = parser.parse_args(["--major-version",
                                         major_version])

        response = self._versions_response(major_version)
        self.requests_mock.register_uri(
            'GET',
            os.path.join(self.url,
                         'vnflcm/v{}/api_versions'.format(major_version)),
            json=response, headers=self.header)

        colmns, data = self.vnflcm_versions.take_action(parsed_args)

        self.assertEqual(colmns, tuple(response.keys()))
        self.assertEqual(data, tuple(response.values()))



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from requests_mock.contrib import fixture as requests_mock_fixture
import testtools
from unittest import mock

from cliff import columns as cliff_columns


class FixturedTestCase(testtools.TestCase):
    client_fixture_class = None
    api_version = '1'

    def setUp(self):
        super(FixturedTestCase, self).setUp()
        self.app = mock.MagicMock()
        if self.client_fixture_class:
            self.requests_mock = self.useFixture(requests_mock_fixture.
                                                 Fixture())
            fix = self.client_fixture_class(self.requests_mock,
                                            api_version=self.api_version)
            self.cs = self.useFixture(fix).client

    def check_parser(self, cmd, args, verify_args):
        cmd_parser = cmd.get_parser('check_parser')
        try:
            parsed_args = cmd_parser.parse_args(args)
        except SystemExit:
            raise ParserException
        for av in verify_args:
            attr, value = av
            if attr:
                self.assertIn(attr, parsed_args)
                self.assertEqual(getattr(parsed_args, attr), value)
        return parsed_args

    def assertNotCalled(self, m, msg=None):
        """Assert a function was not called"""

        if m.called:
            if not msg:
                msg = 'method %s should not have been called' % m
            self.fail(msg)

    def assertListItemsEqual(self, expected, actual):
        """Assertion based on human_readable values of list items"""

        self.assertEqual(len(expected), len(actual))
        for col_expected, col_actual in zip(expected, actual):
            if isinstance(col_actual, tuple):
                self.assertListItemsEqual(col_expected, col_actual)
            elif isinstance(col_expected, cliff_columns.FormattableColumn):
                self.assertIsInstance(col_actual, col_expected.__class__)
                self.assertEqual(col_expected.human_readable(),
                                 col_actual.human_readable())
            else:
                self.assertEqual(col_expected, col_actual)


class ParserException(Exception):
    pass



# Copyright (C) 2022 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from osc_lib.command import command
from osc_lib import utils
from tackerclient.common import exceptions
from tackerclient.i18n import _
from tackerclient.osc import sdk_utils
from tackerclient.osc import utils as tacker_osc_utils

LOG = logging.getLogger(__name__)

_LCCN_SUBSCRIPTION_ID = 'subscription_id'

_MIXED_CASE_FIELDS = ['filter', 'callbackUri']

_FORMATTERS = {
    'filter': tacker_osc_utils.FormatComplexDataColumn,
    '_links': tacker_osc_utils.FormatComplexDataColumn
}


def _get_columns(lccn_subsc_obj):

    column_map = {
        'id': 'ID',
        'filter': 'Filter',
        'callbackUri': 'Callback URI',
        '_links': 'Links'
    }

    return sdk_utils.get_osc_show_columns_for_sdk_resource(lccn_subsc_obj,
                                                           column_map)


class CreateLccnSubscription(command.ShowOne):
    _description = _("Create a new Lccn Subscription")

    def get_parser(self, prog_name):
        parser = super(CreateLccnSubscription, self).get_parser(prog_name)
        parser.add_argument(
            'create_request_file',
            metavar="<param-file>",
            help=_('Specify create request parameters in a json file.'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        subsc = client.create_lccn_subscription(
            tacker_osc_utils.jsonfile2body(parsed_args.create_request_file))
        display_columns, columns = _get_columns(subsc)
        data = utils.get_item_properties(sdk_utils.DictModel(subsc),
                                         columns, formatters=_FORMATTERS,
                                         mixed_case_fields=_MIXED_CASE_FIELDS)
        return (display_columns, data)


class DeleteLccnSubscription(command.Command):
    _description = _("Delete Lccn Subscription(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteLccnSubscription, self).get_parser(prog_name)
        parser.add_argument(
            _LCCN_SUBSCRIPTION_ID,
            metavar="<subscription-id>",
            nargs="+",
            help=_("Lccn Subscription ID(s) to delete"))
        return parser

    def take_action(self, parsed_args):
        error_count = 0
        client = self.app.client_manager.tackerclient
        lccn_subscriptions = parsed_args.subscription_id
        for lccn_subscription in lccn_subscriptions:
            try:
                client.delete_lccn_subscription(lccn_subscription)
            except Exception as e:
                error_count += 1
                LOG.error(_("Failed to delete Lccn Subscription with "
                            "ID '%(subsc)s': %(e)s"),
                          {'subsc': lccn_subscription, 'e': e})

        total = len(lccn_subscriptions)
        if (error_count > 0):
            msg = (_("Failed to delete %(error_count)s of %(total)s "
                     "Lccn Subscriptions.") % {'error_count': error_count,
                                               'total': total})
            raise exceptions.CommandError(message=msg)
        else:
            if total > 1:
                print(_('All specified Lccn Subscriptions are deleted '
                        'successfully'))
            else:
                print(_("Lccn Subscription '%s' is deleted "
                        "successfully") % lccn_subscriptions[0])


class ListLccnSubscription(command.Lister):
    _description = _("List Lccn Subscriptions")

    def get_parser(self, program_name):
        parser = super(ListLccnSubscription, self).get_parser(program_name)
        parser.add_argument(
            "--filter",
            metavar="<filter>",
            help=_("Attribute-based-filtering parameters"),
        )
        return parser

    def get_attributes(self, exclude=None):
        fields = [
            {
                "key": "id",
                "value": "ID"
            },
            {
                "key": "callbackUri",
                "value": "Callback URI"
            }
        ]

        attributes = []

        for field in fields:
            attributes.extend([(field['key'], field['value'],
                              tacker_osc_utils.LIST_BOTH)])
        return tuple(attributes)

    def take_action(self, parsed_args):
        params = {}

        if parsed_args.filter:
            params['filter'] = parsed_args.filter

        client = self.app.client_manager.tackerclient
        subscriptions = client.list_lccn_subscriptions(**params)
        headers, columns = tacker_osc_utils.get_column_definitions(
            self.get_attributes(), long_listing=True)

        dictionary_properties = (utils.get_dict_properties(
            s, columns, mixed_case_fields=_MIXED_CASE_FIELDS)
            for s in subscriptions
        )

        return (headers, dictionary_properties)


class ShowLccnSubscription(command.ShowOne):
    _description = _("Display Lccn Subscription details")

    def get_parser(self, program_name):
        parser = super(ShowLccnSubscription, self).get_parser(program_name)
        parser.add_argument(
            _LCCN_SUBSCRIPTION_ID,
            metavar="<subscription-id>",
            help=_('Lccn Subscription ID to display'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        obj = client.show_lccn_subscription(parsed_args.subscription_id)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj),
            columns, formatters=_FORMATTERS,
            mixed_case_fields=_MIXED_CASE_FIELDS)
        return (display_columns, data)
